!--------------------------------------------c
!  Example: simple data type                 c
!  Send and Receive an Integer Data          c
!  Send and Receive a  Double Precision Data c
!  Send and Receive an Array of Integers     c
!--------------------------------------------c


      program main

      use mpi
      implicit none

      integer ierr,rank,size

      integer i,j
      integer status(MPI_STATUS_SIZE)


!--- data to communicate------------------------------------------------------
      integer MSIZE
      parameter (MSIZE=10)
      integer data_int
      double precision data_double
      double precision matrix(MSIZE,MSIZE)



!--- Start up MPI  ----------- -----------------------------------------------

      call MPI_INIT(ierr)
      call MPI_COMM_RANK(MPI_COMM_WORLD,rank,ierr)
      call MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)

!--- Check number of processes ----------------------------------------------

      if (size.lt.2) then
         if (rank.eq.0) print *, 'SORRY: need at least 2 processes'

!--- exit MPI on error ------------------------------------------------------
         call MPI_FINALIZE(ierr)
         stop

      endif

!--- INTEGER TYPE -----------------------------------------------------------

      if (rank.eq.0) then

         data_int = 10

         CALL MPI_SEND(data_int, 1, MPI_INTEGER, 1, 666, &
               MPI_COMM_WORLD, ierr)
      end if
      if (rank.eq.1) then

         CALL MPI_RECV(data_int, 1, MPI_INTEGER, 0, 666, &
               MPI_COMM_WORLD, status, ierr)

         print *,'Proc 1 receives ',data_int,'from proc 0'
         print *,' '
      endif

!---  ARRAY OF DOUBLE TYPE  --------------------------------------------------

      if (rank.eq.0) then

         do i=1,MSIZE
            do j=1,MSIZE
               matrix(i,j)= dble(i+j)
            enddo
         enddo

         CALL MPI_SEND(matrix, MSIZE*MSIZE, MPI_DOUBLE_PRECISION, 1, &
             666,MPI_COMM_WORLD, ierr)
      end if

      if (rank.eq.1) then

         CALL MPI_RECV(matrix, MSIZE*MSIZE, MPI_DOUBLE_PRECISION, 0, &
             666,MPI_COMM_WORLD, status, ierr)

         print *,'Proc 1 receives the following matrix from proc 0'
         write (*,'(10(f6.2,2x))') matrix
      endif

!--- QUIT --------------------------------------------------------------------
      call MPI_FINALIZE(ierr)
            
      end
