!---------------------------------------------c
!  Exercise: Pi                               c
!                                             c
!  Compute the value of PI using the integral c
!  pi = 4* int 1/(1+x*x)    x in [0-1]        c
!                                             c
!  The integral is approximated by a sum of   c
!  n interval.                                c
!                                             c
!  The approximation to the integral in each  c
!  interval is: (1/n)*4/(1+x*x).              c
!                                             c
!  Each process then adds up every            c
!  n'th interval:                             c
!  (x = rank/n, rank/n+size/n,...)  c
!                                             c
!  The sums computed by each process are      c
!  added together                             c
!---------------------------------------------c



      program main

      use mpi
      implicit none


      integer ierr,rank,size
      integer tag,from,to
      integer i
      integer status(MPI_STATUS_SIZE)

      integer intervals
      double precision dx,sum,x
      double precision f,pi
      double precision in

      double precision  PI25DT
      parameter        (PI25DT = 3.141592653589793238462643d0)



!--- Executive Stantments ---------------------------


!--- Start up MPI ------------------------------------------------

      call MPI_INIT(ierr)
      call MPI_COMM_RANK(MPI_COMM_WORLD,rank,ierr)
      call MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)

      intervals=1000
      sum=0.d0
      dx=1.d0/dble(intervals)

      if(rank.eq.0) then
       write(*,*) 'Evaluation of pi example: each process evaluates'
       write(*,*) 'a subsection of the integral 4/(1+x*x) between 0'
       write(*,*) 'and 1,then sends its partial result to the master.'
       write(*,*)
      endif
!--- Each Process computes integral ------------------------------
      do i=rank+1,intervals,size
         x=dx*(dble(i)-0.5d0)
         f=4.d0/(1.d0+x*x)
         sum=sum+f
      enddo
      
      pi=dx*sum
 
!---  slave processes send partial pi sum to master process 0 ----
!--- Arbitraril choose 201 to our tag ----------------------------

      tag=201
      to=0

      if(rank.ne.0) then
         CALL MPI_SEND(pi,1, MPI_DOUBLE_PRECISION, to,tag, &
              MPI_COMM_WORLD,ierr)
      else
         do from=1,size-1
            print '(a11,2x,f20.10)','I have pi =',pi
            CALL MPI_RECV(in, 1, MPI_DOUBLE_PRECISION, from, tag, &
                 MPI_COMM_WORLD,status, ierr)
            print '(a11,2x,f20.10,2x,a9,i4)','.. received',in &
                  ,'from proc',from
            pi=pi+in
         enddo
      endif
    

      if(rank.eq.0) then
         PRINT '(a40)','***************************************'
         PRINT '(a13,2x,f30.25)','Computed PI =', pi
         PRINT '(a13,2x,f30.25)','The True PI =', PI25DT
      endif
         

!--- QUIT  --------------------------------------------------------------------
      call MPI_FINALIZE(ierr)
      
      end
      
