!------------------------------------------c
! Exercise: Ring
!
! In this exercise you will communicate an integer
! among process in a circular topology, so that
! every process has a left and a right neighbor
! to receive and to send data respectivelly.
!
! Please, read carefully the comments before the
! do while cycle to understand the rules of the game.
!
!------------------------------------------c

      program main

      use mpi
      implicit none

      integer ierr,rank,size
      integer tag,from,next

      integer num

      integer status(MPI_STATUS_SIZE)
      
!--- Start up MPI  ----------- ----------------------------------------------
      call MPI_INIT(ierr)
      call MPI_COMM_RANK(MPI_COMM_WORLD,rank,ierr)
      call MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)

!---------------------------------------------------------------------------c
!    Arbitrarily choose 201 to be our tag. Calculate the rank of the        c
!    next process. Use the modulus operator so that the                     c
!    last process "wraps around" to rank zero.                              c
!---------------------------------------------------------------------------c

      tag=201
      next = mod(rank + 1,size)
      from = mod(rank + size - 1,size)

      num=1
      
!--- If we are the console process, get an integer from the user to    
!--- specify how many times we want to go around the ring

      if(rank.eq.0) then
         print '(a50)','Enter the number of times around the ring'
         read(*,*) num
         if(num.lt.0) num=1

         print '(a,2x,i4,2x,a,2x,i4,2x,a,2x,i4)','Proc ',rank, &
              ' sends a bag with ',num,' sandwich to proc',next

         CALL MPI_SEND(num,1, MPI_INTEGER,next,tag,MPI_COMM_WORLD &
             ,ierr)
         
      endif

!--- Main process start the ring --------------------------------------


!----------------------------------------------------------------------------c
! Pass the message around the ring.  The exit mechanism works as             c
! follows: the message (a positive integer) is passed around the             c
! ring.  Each time it passes rank 0, it is decremented.  When each           c
! processes receives the 0 message, it passes it on to the next              c
! process and then quits.  By passing the 0 first, every process             c
! gets the 0 message and can quit normally.                                  c
!----------------------------------------------------------------------------c



      do while(num.gt.0) 
         CALL MPI_RECV(num,1, MPI_INTEGER, from, tag,MPI_COMM_WORLD, &
             status, ierr)
         print '(a,2x,i4,2x,a,2x,i4,2x,a)','Proc ',rank, &
              ' receives a bag with ',num,' sandwich'
         if(rank.eq.0) then
            num=num-1
            print '(a40)','* * * Process 0 eats one sandwich * * *'
         endif
         print '(a,2x,i4,2x,a,2x,i4,2x,a,2x,i4)','Proc ',rank, &
             ' sends a bag with ',num,' sandwich to proc ',next
         
         CALL MPI_SEND(num,1, MPI_INTEGER,next,tag,MPI_COMM_WORLD &
             ,ierr)
         
      enddo
      

!--- End of the ring ------------------------------------------------

      print '(a,2x,i4,2x,a)','Proc ',rank,' leaves the ring'
      print *,'*******************************************'

!--- The last process send 0 to process 0, which needs to
!--- be received before the program can exit 


      if(rank.eq.0) then
         CALL MPI_RECV(num,1, MPI_INTEGER, from, tag,MPI_COMM_WORLD, &
             status, ierr)
         print '(a)','...process 0 starves'
      endif

!--- QUIT -------------------------------------------------------------------
      
      call MPI_FINALIZE(ierr)
      
      end
