!--------------------------------------------c
!  Exercise: timing blocking sends
!
!  In this exercise you will try blocking
!  communications (syncronous and buffered)
!  to send a matrix from process 0 to process 1.
!
! Follow this rules:
!  - Before the two communication, a barrier should
!    syncronize the processes;
!  - Process 0 will begin the blocking send
!  - while process 1 will sleep for 5 seconds 
!    before starting the receive
!  - measure how long process 0 will take in
!    the two different sending calls
!--------------------------------------------c

program main

      use mpi

      implicit none

      integer ierr,rank,size

      integer i,j
      integer status(MPI_STATUS_SIZE)


!--- data to communicate------------------------------------------------------
      integer MSIZE
      parameter (MSIZE=10000)

      double precision matrix(MSIZE,MSIZE)


      INTEGER mpibuffer_length, typesize
      ! timing
      double precision startTime, endTime
      double precision, DIMENSION(:), ALLOCATABLE :: mpibuffer
      


!--- Start up MPI  ----------- -----------------------------------------------

      call MPI_INIT(ierr)
      call MPI_COMM_RANK(MPI_COMM_WORLD,rank,ierr)
      call MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)

!--- Check number of processes ----------------------------------------------

      if(size.ne.2) then
         if(rank.eq.0) print *, 'I need two processes to run'

!--- exit MPI on error ------------------------------------------------------
         call MPI_FINALIZE(ierr)
         stop

      endif

!---  Synchronous communication -----------------------------------------


      if(rank.eq.0) then
        print *, 'Communication mode examples: process 0 sends a vector'
        print *, 'to process 1 using a synchronous and a buffered send.'
        print *, 'Process 1 waits for 5 seconds before starting the'
        print *, 'corresponding receive call. Time spent by process 0'
        print *, 'in the synchronous and buffered sends is reported.'

         do i=1,MSIZE
            do j=1,MSIZE
               matrix(i,j)= dble(i+j)
            enddo
         enddo

      endif

      CALL MPI_BARRIER(MPI_COMM_WORLD, ierr)

      if(rank.eq.0) then
         startTime = MPI_Wtime()

         CALL MPI_SSEND(matrix, MSIZE*MSIZE, MPI_DOUBLE_PRECISION, 1, &
              666,MPI_COMM_WORLD, ierr)

         endTime = MPI_Wtime()
         
         print '(a12,2x,f10.7,2x,a18)','Proc 0 spent',endTime - startTime &
              ,'to exit from SSend.'

      else

!---  Process 1 sleeps for 5 seconds and receive ---------------------
!         CALL sleep(%VAL(5))
         CALL sleep(5)
!        call system( 'sleep 5' )
         CALL MPI_RECV(matrix, MSIZE*MSIZE, MPI_DOUBLE_PRECISION, 0, &
            666, MPI_COMM_WORLD, status, ierr)

      endif


!---  Buffered Communications ------------------------------------------

      if(rank.eq.0) then
         
         do i=1,MSIZE
            do j=1,MSIZE
               matrix(i,j)= dble(i+j)
            enddo
         enddo
         
      endif

      CALL MPI_BARRIER(MPI_COMM_WORLD, ierr)

      if(rank.eq.0) then

         ALLOCATE(mpibuffer(msize*msize))
         CALL MPI_TYPE_SIZE(MPI_DOUBLE_PRECISION, typesize, ierr)
         mpibuffer_length=typesize*msize*msize+MPI_BSEND_OVERHEAD

         CALL MPI_BUFFER_ATTACH(mpibuffer,mpibuffer_length,ierr)


         startTime = MPI_Wtime()

         CALL MPI_BSEND(matrix,MSIZE*MSIZE,MPI_DOUBLE_PRECISION,1,666 &
              ,MPI_COMM_WORLD,ierr)

         endTime = MPI_Wtime()

        CALL MPI_BUFFER_DETACH(mpibuffer,mpibuffer_length,ierr)
         print '(a12,2x,f10.7,2x,a18)','Proc 0 spent',endTime - startTime &
                ,'to exit from BSend.'


      else

!---  Process 1 sleeps for 5 seconds and receive ---------------------
!         CALL sleep(%VAL(5))
         CALL sleep(5)
!        call system( 'sleep 5' )
         CALL MPI_RECV(matrix, MSIZE*MSIZE, MPI_DOUBLE_PRECISION, 0, &
            666, MPI_COMM_WORLD, status, ierr)
         
      endif


!--- QUIT --------------------------------------------------------------------
      call MPI_FINALIZE(ierr)
            
end program main
