#include <stdio.h>
#include <mpi.h>

/* MSIZE is the size of the integer arrays to send */
#define MSIZE 10000

int main(int argc, char *argv[])
{
    MPI_Status status;
    int rank, size, tag;

    int i;
    int A[MSIZE], B[MSIZE];

    /* cartesian topology */
    int periods[] = {1}, dims[]={0};
    MPI_Comm comm_cart;
    int source, dest;

    /* Start up MPI */
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);


    for (i = 0; i < MSIZE; i++)
        A[i] = rank;
    tag = 201;

    MPI_Dims_create(size, 1, dims);
    MPI_Cart_create(MPI_COMM_WORLD, 1, dims, periods, 0, &comm_cart);

    MPI_Cart_shift(comm_cart, 0, 1, &source, &dest);
    MPI_Sendrecv(A, MSIZE, MPI_INT, dest, tag, B, MSIZE, MPI_INT, source, 
		    tag, comm_cart, &status);


    printf("Proc %d sents %d integers to proc %d\n",
        rank, MSIZE, dest);
    printf("Proc %d receives %d integers from proc %d\n",
        rank, MSIZE, source);

    /* print first content of arrays A and B */
    printf("Proc %d has A[0] = %d, B[0] = %d\n", rank, A[0], B[0]);

    /* Quit */
    MPI_Finalize();

    return 0;
}

