#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>


int main(int argc, char *argv[])
{
    MPI_Status status;
    int rank, size, tag, tag1;
    int rank_cart;
    int source, dest;
    int i, j;
    int value, north, south, east, west;
    double average, toBePrinted;

    /* cartesian topology */
    MPI_Comm comm_cart;
    int periods[] = {1, 1};
    int dims[2]={0,0}, coords[2];

    /* Start up MPI */
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);


    value = rank;

    MPI_Dims_create(size, 2, dims);
    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, 0, &comm_cart);

    tag = 200;
    tag1 = 201;

    /* to nord/from south */
    MPI_Cart_shift(comm_cart, 0, 1, &source, &dest);
    MPI_Sendrecv(&value, 1, MPI_INT, dest, tag, &south, 1, MPI_INT, source, 
		    tag, comm_cart, &status);
    /* to south/from north */
    MPI_Cart_shift(comm_cart, 0, -1, &source, &dest);
    MPI_Sendrecv(&value, 1, MPI_INT, dest, tag1, &north, 1, MPI_INT, source, 
		    tag1, comm_cart, &status);
    /* to east/from west */
    MPI_Cart_shift(comm_cart, 1, 1, &source, &dest);
    MPI_Sendrecv(&value, 1, MPI_INT, dest, tag, &west, 1, MPI_INT, source, 
		    tag, comm_cart, &status);
    /* to west/from east */
    MPI_Cart_shift(comm_cart, 1, -1, &source, &dest);
    MPI_Sendrecv(&value, 1, MPI_INT, dest, tag1, &east, 1, MPI_INT, source, 
		    tag1, comm_cart, &status);

    average = 0.25 * (north + south + east + west);

    /* writing results */
    if (rank == 0)
    {
       MPI_Comm_rank(comm_cart, &rank_cart);
       for (i=0; i < dims[0]; i++)
       {
          for (j=0; j < dims[1]; j++)
          {
              coords[0] = i;
              coords[1] = j;
              MPI_Cart_rank(comm_cart, coords, &source);
              if (source != rank_cart)
              {
                 MPI_Recv( &toBePrinted, 1, MPI_DOUBLE, source, tag, comm_cart, &status);
                 printf(" %d:(%d,%d) %5.2lf     ", source, i, j, toBePrinted);
              }
              else
                 printf(" %d:(%d,%d) %5.2lf     ", source, i, j, average);

          }
          printf("\n");
       }
    }
    else
       MPI_Send(&average, 1, MPI_DOUBLE, 0, tag, comm_cart);


    /* Quit */
    MPI_Finalize();

    return 0;
}

